sc;; O functie foarte utila este build-list. Aceasta primeste un numar n si o functie unara f si intoarce
;; lista formata prin aplicarea functiei f asupra fiecarui numar de la 0 la n - 1
;; Exemple:
;; (build-list 10 (λ(x) x)) ; -> '(0 1 2 3 4 5 6 7 8 9)
;; (build-list 10 (λ(x) (* x x))) ; -> (0 1 4 9 16 25 36 49 64 81)
;; (build-list 5 (λ(x) 0)) ; -> '(0 0 0 0 0)

;; Functii curry si uncurry
;;--------------------------

;; 1. (0.5p)
;; sa se implementeze functia uncurry->curry, dupa urmatoarea specificatie:
;; fiind data o functie uncurry f care ia in Scheme 2 argumente,
;; (uncurry->curry f) trebuie sa produca o functie curry cu acelasi efect ca functia f
;; practic uncurry->curry este o functie care primeste mai intai ca parametru o functie
;; uncurry, apoi argumentele acesteia (pe rand, pt ca altfel (uncurry->curry f) nu ar 
;; fi o functie curry) si face exact ce facea f asupra acelor argumente
(define uncurry->curry
  (lambda (f)
    (lambda (x)
      (lambda (y)
        (f x y)))))

(((uncurry->curry >) 5) 4) ;=> #t

;; 2. (0.5p)
;; sa se implementeze functia curry->uncurry (procesul invers celui de la punctul 1)
;; acum (curry->uncurry g) va produce o functie uncurry dintr-o functie curry g de 
;; 2 argumente
(define curry->uncurry
  (λ (f)
    (λ (x y)
      ((f x) y))))


((curry->uncurry (uncurry->curry >)) 5 4) ;=> #t

;; Functionale
;;------------

;; functionala = o functie care primeste functii ca parametri sau returneaza functii
;; in specificatiile de mai jos
;; f = functie binara
;; p = predicat (functie care testeaza o proprietate a argumentului
;; sau si intoarce true sau false)
;; L, L1, L2 = liste
;; x = element in lista


;; 3. (1p)
;; sa se implementeze functionala foldright
;; (foldright f init lst) acumuleaza aplicatia functiei binare f pe lista lst in felul urmator:
;; mai intai (f last init) => rezultat1 [last este ultimul element din lst]
;; apoi (f last1 rezultat1) => rezultat2, (f last2 rezultat2) => rezultat3 etc
;; [last1 e penultimul element, last2 cel dinaintea-i]
;; pana la (f (car lst) rezultat-pana-acum)
(define foldright
  (λ (f init lst)
    [if (null? lst)
        init
        (f [car lst] (foldright f init [cdr lst]))]))

(foldright cons '() '(1 2 3 4))
;;4. (1p)
;;Sa se implementeze functionala map folosind foldright
;;(my-map f L) aplica f asupra fiecarui element din L si returneaza lista rezultatelor
(define my-map
  (λ (f L)
    (foldright [λ (e ac) (cons (f e) ac)] '() L)))

(my-map (λ (e) (* 2 e)) '(1 2 3 4 5)) ;; => '(2 4 6 8 10)


;;5. (1p)
;;Sa se implementeze functionala filter folosind foldright
;;(my-filter p L) returneaza lista elementelor din L pentru care predicatul p returneaza #t
(define my-filter
  (λ (p L)
    (foldright [λ (e ac) (if [p e] (cons e ac) ac)] '() L)))

(my-filter (λ (l) (not (null? l))) '((1 2 3) () (7 6 5) () () (9) ())) ;;=> '((1 2 3) (7 6 5) (9))


;;Matrice
;;-------

;; Utilizati functionale pentru a implementa cat mai pe scurt functiile urmatoare.
;; matrice = lista de liste. Numar de linii = (length m). Numar de coloane = (length (car l))
;;
;; Pentru o parte din exercitiile urmatoare ar putea fi de ajutor urmatoarele functii:
;; I_n, care primeste un numar n si intoarce matricea unitate n x n
;; Observati implementarea scurta utilizand functionale.
(define (I_n n)
  (map (λ(k) (build-list n (λ(x) (if (= x k) 1 0)))) (build-list n (λ(x) x))))
;; (I_n 3) ; -> ((1 0 0) (0 1 0) (0 0 1))

;; O_mn, care primeste doua numere m si n si intoarce o matrice plina de 0 de dimensiune m x n
(define (O_mn m n)
  (build-list m (λ(l) (build-list n (λ(e) 0)))))
;; (O_mn 3 4) ; -> ((0 0 0 0) (0 0 0 0) (0 0 0 0))

;; num-mul-matrix primeste o valoare k si o matrice M si inmulteste valoarea cu matricea M
(define (num-mul-matrix k M)
  (map (λ(L) (map (λ(e) (* e k)) L)) M))
;; (num-mul-matrix 10 '((1 2 3) (4 5 6))) ;-> '((10 20 30) (40 50 60))


;;6. (0.5p)
;;Sa se implementeze functia get-line
;;(get-line l M) returneaza a l-a linie a matricei M. Liniile sunt indexate de la 1
;;Hint: linia l este al l-lea element din lista.
(define get-line
  (λ (l M)
    (cond
      [(or (null? M) (< l 1)) '()]
      [(= l 1) [car M]]
      [else (get-line (- l 1) [cdr M])])))

(get-line 2 '((1 2 3 4) (5 6 7 8))) ;;=> '(5 6 7 8)


;;7. (0.5p)
;;Sa se implementeze functia get-column
;;(get-column c M) returneaza a c-a coloana a matricei M. Coloanele sunt indexate de la 1.
;;Hint: coloana c contine al c-lea element din fiecare linie

(define get-elem get-line)

(define get-column
  (λ (c M)
    (map [λ (L) (get-elem c L)] M)))

(get-column 2 '((1 2) (3 4) (5 6) (7 8))) ;;=> '(2 4 6 8)


;;8. (2p)
;;Sa se implementeze functia transpose folosind cel putin o functionala
;;(transpose M) returneaza transpusa matricei M
(define transpose
  (λ (M)
    (foldright [λ (L ac) (map cons L ac)] (build-list (length [car M]) [λ (e) '()]) M)))

;;sau
;(define transpose
;  (λ (M)
;    (apply map list M)))

(transpose '((1 2 3) (4 5 6))) ;;=> '((1 4) (2 5) (3 6)) 

;;9. (1p)
;;Sa se implementeze functia add-matrix folosind cel putin o functionala
;;(add-matrix M1 M2) returneaza suma matricelor M1 si M2.
(define add-matrix
  (λ (M1 M2)
    (map [λ (L1 L2) (map + L1 L2)] M1 M2)))

(add-matrix '((1 2 3 4) (6 7 8 9)) '((1 2 2 1) (0 1 3 5))) ;;=> '((2 4 5 5) (6 8 11 14)) 


;;10. (1p)
;;Sa se implementeze functia mul-matrix folosind cel putin o functionala
;;(mul-matrix M1 M2) returneaza produsul matricelor M1 si M2.


;(define mul-matrix
;  (λ (M1 M2)
;    (map [λ (L)
;           (map [λ (C) (foldr [λ (e1 e2 init) (+ init (* e1 e2))] 0 L C)] (transpose M2))] M1)))

;;sau, considerând (mul-matrix M1 M2) = (transpose (mul-matrix (transpose M2) (transpose M1))),
;;se fac doar două operații de transpunere, în loc de (length M1) operații
(define mul-matrix
  (λ (M1 M2)
    [transpose (map [λ (L) 
                      (map [λ (C) (foldr [λ (e1 e2 init) (+ init (* e1 e2))] 0 L C)] M1)] (transpose M2))]))

(mul-matrix '((1 2) (3 4) (5 6)) '((7 8 9) (10 11 12))) ;;=> '((27 30 33) (61 68 75) (95 106 117))


;;11. (1p)
;;Sa se implementeze functia pow-matrix folosind cel putin o functionala
;;(pow-matrix M p) returneaza rezultatul ridicarii matricei M la puterea p. p este natural.
(define pow-matrix
  (λ (M p)
    (foldright [λ (e init) (mul-matrix M init)] (I_n (length M)) (build-list p [λ (e) e]))))

;;;sau
;(define pow-matrix
;  (λ (M p)
;    (foldright mul-matrix (I_n (length M)) (build-list (length M) [λ (e) M]))))


(pow-matrix '((1 2) (3 4)) 2) ;;=> '((7 10) (15 22))


;;12. Bonus (2p)
;;Sa se implementeze functia exp-matrix
;;((exp-matrix n) M) returneaza aproximarea e^M (unde M este o matrice) folosind primii n termeni din seria Taylor.
;;Utilizati urmatoarea formula: e^M = (1/0!) * M^0 + (1/1!) * M^1 + (1/2!) * M^2 + ... (1/(n-1)!) * M^(n - 1)

;(define factorial
;  (λ (n)
;    (if (= n 0)
;        1
;        (* n (factorial (- n 1))))))
;
;(define exp-matrix
;  (λ (n)
;    (λ (M)
;      (foldl [λ (f Mp init) (add-matrix (num-mul-matrix (/ 1.0 f) Mp) init)]
;             (O_mn (length M) (length M))
;             (build-list n [λ (e) (factorial e)])
;             (build-list n [λ(e) (pow-matrix M e)])))))

;;sau, dacă observăm că lista de factoriale se contruieste identic cu lista de puteri ale unei matrice,
;;putem construi o functionala, unfold [vezi descriere]

(define unfold-iter
  (λ (n f elem acum i)
    (if (= i n)
        (reverse acum)
        (unfold-iter n f elem (cons (f (elem i) [car acum]) acum) (+ i 1)))))

;;unfold: aplica o functie pe un element si init la prima iteratie.
;;La următoarele iterații, adaugă la lista de rezultate valoarea obtinuta in urma 
;;aplicarii functiei f pe un nou element si rezultatul iteratiei precedente
;;n: numarul de iteratii
;;f: functie cu doua argumente, un element si ultimul rezultat [sau un element si valoarea initiala]
;;elem: functie care calculeaza elementul nou. Este aplicata pe [0, (- n 1)], ca in build-list
;;init: valoarea initiala
(define unfold
  (λ (n f elem init)
    (unfold-iter n f elem (list init) 0)))

(define fact-list
  (λ (n)
    (unfold n * [λ (e) (+ e 1)] 1)))

(define pow-mat-list
  (λ (n M)
    (unfold n mul-matrix [λ (e) M] (I_n (length M)))))
        
(define exp-matrix
  (λ (n)
    (λ (M)
      (foldl [λ (f Mp init) (add-matrix (num-mul-matrix (/ 1.0 f) Mp) init)] (O_mn (length M) (length M)) (fact-list n) (pow-mat-list n M)))))

((exp-matrix 20) '((1 2) (3 4))) ;;=> '((51.96 74.73) (112.10 164.07))